/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.maven;

// REMPL API from com.rempl:rempl-api
import com.rempl.api.ROM;
import com.rempl.api.ROMException;
import com.rempl.api.Rempl;
import com.rempl.api.RemplFactory;
import com.rempl.api.Rempler;

// for file management
import java.io.File;

// for file management
import java.net.URL;

// java basic classes
import java.util.List;
import java.util.Properties;

// regular expressions
import java.util.regex.Matcher;
import java.util.regex.Pattern;

// file manipulations
import org.apache.commons.io.FileUtils;

// MOJO API
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;

// SLF4J logger
import org.slf4j.impl.StaticLoggerBinder;

// Aether API
import org.sonatype.aether.RepositorySystem;
import org.sonatype.aether.RepositorySystemSession;
import org.sonatype.aether.repository.RemoteRepository;

/**
 * Abstract MOJO that requests dependency injection from Maven.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: AbstractRemplMojo.java 1380 2011-05-10 19:51:08Z guard $
 * @todo #15 "Aggregate" option has to be enabled, in order to
 *           collect information from sub-modules (read from them)
 *           and then aggregate UML models together and run
 *           reporting only once. This aggregator has to understand
 *           that it has to be executed only once. Somehow...
 *           @see http://stackoverflow.com/questions/5555699
 * @todo #31 The class definitely needs refactoring since now it's too
 *       big and has too many methods. I suppressed PMD warning about
 *       it at the moment, but some refactoring is required in the future.
 */
@SuppressWarnings("PMD.TooManyMethods")
public abstract class AbstractRemplMojo extends AbstractMojo {

    /**
     * The entry point to Aether, i.e. the component
     * doing all the work.
     * @component
     */
    private RepositorySystem system;

    /**
     * The current repository/network configuration of Maven.
     * @parameter expression="${repositorySystemSession}"
     * @readonly
     */
    private RepositorySystemSession session;

    /**
     * The project's remote repositories to use for the resolution
     * of plugins and their dependencies.
     * @parameter expression="${project.remoteProjectRepositories}"
     * @readonly
     */
    private List<RemoteRepository> remotes;

    /**
     * The <tt>rom.xml</tt> file location.
     * @parameter expression="${rempl.rom}" default-value="${basedir}/src/main/rempl/rom.xml" property="rom"
     * @required
     */
    private String romName;

    /**
     * The XMI file.
     * @parameter expression="${rempl.xmi}" default-value="${project.build.directory}/rempl.xmi" property="xmi"
     * @required
     */
    private File xmiFile;

    /**
     * Skip everything.
     * @parameter expression="${rempl.skip}" default-value="false" property="skip"
     * @required
     */
    private boolean skip;

    /**
     * Shall we ignore this report if the ROM.xml file
     * is absent?
     * @parameter default-value="false" property="ignoreIfAbsent"
     * @required
     */
    private boolean ignoreIfAbsent;

    /**
     * Maven project, to be injected by Maven itself.
     * @parameter expression="${project}"
     * @required
     */
    private MavenProject project;

    /**
     * Public ctor.
     */
    public AbstractRemplMojo() {
        super();
        this.xmiFile = null;
        this.romName = null;
    }

    /**
     * Inject repository system.
     * @param sys The system
     */
    public final void setSystem(final RepositorySystem sys) {
        this.system = sys;
    }

    /**
     * Inject repository system session.
     * @param ses The system session
     */
    public final void setSession(final RepositorySystemSession ses) {
        this.session = ses;
    }

    /**
     * Inject repositories.
     * @param reps The list of repos
     */
    public final void setRemotes(final List<RemoteRepository> reps) {
        this.remotes = reps;
    }

    /**
     * Shall we ignore execution if ROM is absent?
     * @param ignore Shall we ignore?
     */
    public final void setIgnoreIfAbsent(final boolean ignore) {
        this.ignoreIfAbsent = ignore;
    }

    /**
     * Shall we skip execution?
     * @param yes Shall we skip?
     */
    public final void setSkip(final boolean yes) {
        this.skip = yes;
    }

    /**
     * Set Maven Project (used mostly for unit testing).
     * @param proj The project to set
     */
    public final void setProject(final MavenProject proj) {
        this.project = proj;
    }

    /**
     * Get Maven Project (used mostly for unit testing).
     * @return The project
     */
    public final MavenProject getProject() {
        return this.project;
    }

    /**
     * Shall we ignore execution if ROM is absent?
     * @return Shall we ignore?
     */
    public final boolean ignoreIfAbsent() {
        return this.ignoreIfAbsent;
    }

    /**
     * Shall we skip execution?
     * @return Shall we skip?
     */
    public final boolean skip() {
        return this.skip;
    }

    /**
     * Set <tt>rom.xml</tt> location.
     * @param name The location of the rom.xml
     */
    public final void setRom(final String name) {
        this.romName = name;
    }

    /**
     * Set XMI file location.
     * @param pfile The location of the XMI file
     */
    public final void setXmi(final File pfile) {
        this.xmiFile = pfile;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void execute() throws MojoExecutionException {
        StaticLoggerBinder.getSingleton().setMavenLog(this.getLog());
        if (this.skip()) {
            this.getLog().info("Execution skipped");
        } else if (this.romFile() == null && this.ignoreIfAbsent()) {
            this.getLog().info("Ignore since ROM is not found");
        } else {
            this.run();
        }
    }

    /**
     * Execute the MOJO.
     * @throws MojoExecutionException If something goes wrong
     */
    protected abstract void run() throws MojoExecutionException;

    /**
     * Get ROM file location.
     * @return The location
     */
    protected final File romFile() {
        if (this.romName == null) {
            return null;
        }
        File file = new File(this.romName);
        if (!file.exists()) {
            this.getLog().debug("ROM file is absent: " + this.romName);
            // maybe it's a resource?
            final URL url = AbstractRemplMojo.class
                .getResource(this.romName);
            if (url != null) {
                file = FileUtils.toFile(url);
                this.getLog().debug("ROM resource found at: " + url);
            } else {
                file = null;
            }
        }
        return file;
    }

    /**
     * Get ROM instance.
     * @return The ROM
     * @throws MojoExecutionException If something goes wrong
     */
    protected final ROM rom() throws MojoExecutionException {
        final File file = this.romFile();
        if (file == null) {
            throw new MojoExecutionException("ROM not found: " + this.romName);
        }
        ROM rom;
        try {
            rom = new ROM(this.substitute(file));
        } catch (ROMException ex) {
            throw new MojoExecutionException("Invalid ROM: " + file, ex);
        }
        this.getLog().debug("ROM.xml found at: " + file);
        return rom;
    }

    /**
     * Get XMI file to read from.
     * @return The file
     * @throws MojoExecutionException If something goes wrong
     */
    protected final File xmiForReading() throws MojoExecutionException {
        if (!this.xmiFile.exists()) {
            throw new MojoExecutionException("XMI not found " + this.xmiFile);
        }
        this.getLog().info("Reading XMI from: " + this.xmiFile);
        return this.xmiFile;
    }

    /**
     * Get XMI file to write to.
     * @return The file
     * @throws MojoExecutionException If something goes wrong
     */
    protected final File xmiForWriting() throws MojoExecutionException {
        if (this.xmiFile.getParentFile().mkdirs()) {
            this.getLog().debug("Directory created for: " + this.xmiFile);
        }
        try {
            if (!this.xmiFile.exists() && this.xmiFile.createNewFile()) {
                this.getLog().debug("File created: " + this.xmiFile);
            }
        } catch (java.io.IOException ex) {
            throw new MojoExecutionException(
                "Invalid XMI: " + this.xmiFile,
                ex
            );
        }
        this.getLog().info("Writing XMI to: " + this.xmiFile);
        return this.xmiFile;
    }

    /**
     * Get rempler.
     * @return The rempler
     * @throws MojoExecutionException If something goes wrong
     */
    protected final Rempler rempler() throws MojoExecutionException {
        final RemplFactory factory = Rempl.getFactory();
        factory.setLoader(new MavenLoader(this.resolver()));
        return factory.makeRempler(this.rom());
    }

    /**
     * Construct a resolver.
     * @return The resolver
     * @see #execute()
     */
    protected final MavenResolver resolver() {
        return new MavenResolver(
            this.system,
            this.session,
            this.remotes
        );
    }

    /**
     * Substitute variable values.
     * @param file The file to work with
     * @return Content with vars already substituted
     * @see #rom()
     * @throws MojoExecutionException If something goes wrong
     */
    private String substitute(final File file) throws MojoExecutionException {
        String content;
        try {
            content = FileUtils.readFileToString(file);
        } catch (java.io.IOException ex) {
            throw new MojoExecutionException(
                "Failed to read content from ROM file",
                ex
            );
        }
        final StringBuffer buf = new StringBuffer();
        final Matcher matcher = Pattern.compile("\\$\\{([\\w\\.\\-]+)\\}")
            .matcher(content);
        final Properties props = this.project.getProperties();
        while (matcher.find()) {
            final String name = matcher.group(1);
            if (!props.containsKey(name)) {
                throw new MojoExecutionException(
                    "Property '" + name + "' not found in project"
                );
            }
            final Object replacer = props.get(name);
            matcher.appendReplacement(buf, replacer.toString());
            this.getLog().debug(
                matcher.group()
                + " replaced with '"
                + replacer + "' in "
                + file
            );
        }
        matcher.appendTail(buf);
        return buf.toString();
    }

}
